/* FILE: sine.dsp               Copyright (c), Symmetric Research, 1993 1998
 *
 * This program computes a precision sine or cosine wave.  It uses
 * polynomial approximation and accurate angle evaluation to produce
 * values suitable for sending to high resolution D/A converters.
 *
 * Assemble this program with dspasm /s /2 sine.dsp, and then run it in
 * the DSPMON debugger by hand.  The program loops, filling the output
 * array over and over with a sine wave.
 *
 * Note that the angle is accurately evaluated at each sample point.  Do
 * not evaulate the angle with formulas like angle += dtheta because
 * considerable accumulated error will result.
 *
 * The send routine listed here saves the output value in an array.  For
 * actual D/A applications, the output value should be scaled and
 * converted to an integer before sending it to an A/D converter.
 *
 * Note that DTHETA must be calculated to be TWOPI/N to get one full
 * cycle on each pass through the while loop.  When you change N, be
 * sure to also enter the new value for DTHETA.
 *
 * The trig polynomials used here were derived using QR to do an L2 fit
 * on [-PI,PI].  Note that the polynomial coefficients could be easily
 * changed to approximate other waveforms.
 *
 *
 *
 */




<0x000800>

/* Program parameters. */

long
        N               =    0x000100;

dspf
        DTHETA          =    2.454369261e-002,   /* must be = TWOPI/N */
        I               =    0.000000000e+000,

        ONE             =    1.000000000e+000,
        ZERO            =    0.000000000e+000,
        PI              =    3.141592654e+000;






<0x000000>

main {

        register

                n               = r1,
                i               = r2,
                dtheta          = r3,
                one             = r4,
                zero            = r5,
                pi              = r6,
                output          = r7;




        /* Generate a continuous sine wave, cycling from -PI to PI. */

                i = I;
           dtheta = DTHETA;

              one = ONE;
             zero = ZERO;
               pi = PI;


  trig_loop:
                n = *N;
               *i = a0 = *zero;

           output = OUTPUT_ARRAY;   /* pointer passed to send ... */


            while ( n-- > 0 ) {

                    a0 = - *pi + a0 * *dtheta;

                        call trig;
                        call send;

                      *i = a0 = *i + *one;

                    }


        goto trig_loop; /* loop for the next sine wave cycle */



}
send {

        register

                output  = r7;  /* must agree with main ... */


            *output++ = a2 = a2; /* save the value of sine */

        return;

}










/* Trig polynomial evaluation:
 *
 * This function computes
 * 
 *      (a1,a2) = ( cos( a0), sin(a0))    where -PI <= a0 <= PI
 *
 *
 * The return value is accurate to 1.0E-6, almost the full accuracy of
 * a DSP32C single precision float mantissa.
 *
 * By interleaving the evaluation of cos and sin polynomials, both can
 * be computed at the same time with little performance loss, while
 * still using a straightforward Horners rule calculation.
 *
 * Registers r12, r13, and a3 are trashed by this routine.
 *
 *
 */

trig {

        register

                cc      = r12,
                sc      = r13;



        /* Evaluate (cos,sin) with polynomial approximation. */

        a3 = a0*a0;                    /* xx = a3 = theta^2 */

          cc = cos_coeff;
          sc = sin_coeff;

            a1 = *cc++ + a3 * *cc++;      /* cos = cc[4] + xx * cc[5]  */
            a2 = *sc++ + a3 * *sc++;      /* sin = sc[4] + xx * sc[5]  */

              nop;

            a1 = *cc++ + a3 * a1;         /* cos = cc[3] + xx * cos    */
            a2 = *sc++ + a3 * a2;         /* sin = sc[3] + xx * sin    */

              nop;

            a1 = *cc++ + a3 * a1;         /* cos = cc[2] + xx * cos    */
            a2 = *sc++ + a3 * a2;         /* sin = sc[2] + xx * sin    */

              nop;

            a1 = *cc++ + a3 * a1;         /* cos = cc[1] + xx * cos    */
            a2 = *sc++ + a3 * a2;         /* sin = sc[1] + xx * sin    */

              nop;

            a1 = *cc++ + a3 * a1;         /* cos = cc[0] + xx * cos    */
            a2 = *sc++ + a3 * a2;         /* sin = sc[0] + xx * sin    */

              nop;
              nop;


            a2 = a0 * a2;                 /* sin = x * sin             */



        return;  /* result is in (a1,a2) ... */



}
dspf

        cos_coeff[] = {

                -2.21543315e-007,       /* c[5] */
                 2.42466636e-005,       /* c[4] */
                -1.38624043e-003,       /* c[3] */
                 4.16609704e-002,       /* c[2] */
                -4.99995555e-001,       /* c[1] */
                 9.99999445e-001,       /* c[0] */

                },

        sin_coeff[] = {

                -2.05577593e-008,       /* c[5] */
                 2.70480024e-006,       /* c[4] */
                -1.98133653e-004,       /* c[3] */
                 8.33259218e-003,       /* c[2] */
                -1.66665829e-001,       /* c[1] */
                 9.99999732e-001,       /* c[0] */

                };








<0x001000>

/* Output value area:
 *
 * Review the values saved here for accuracy after running the program.
 *
 *
 *
 */

dspf

        OUTPUT_ARRAY[];

