/* FILE: intrupt.dsp            Copyright (c), Symmetric Research, 1993 1998
 *
 * This program demonstrates external interrupts on the DSP32C.
 *
 * DSP32C external interrupts are fairly simple.  As soon as a low level
 * is detected on one of the ireq1,2 pins the processor branches to an
 * interrupt service routine (isr).  The isr can execute any interrupt
 * tasks, and then should terminate with an ireturn to return to the
 * mainline code.  Further interrupts on the same signal are disabled
 * until the ireturn is executed.
 * 
 * Register r22 (IVTP) is dedicated by the DSP32C to point to the
 * interrupt vector table.  This table must be set up before enabling
 * interrupts, and r22 should be pointing at it.  The table has two long
 * word entries per interrupt, and these should be filled with either
 * branches to interrupt service routines, or "quick interrupts"
 * comprised of an instruction and ireturn.  This program demonstrates
 * filling the table in with branches, while the program quick.dsp shows
 * how to use a two instruction sequence.
 * 
 * To run this program, load it by hand and start execution in DSPMON.
 * Check the value of the count memory variable.  Then bring the
 * interrupt pin low and check the count again.  It should be
 * incrementing while the interrupt pin is low and not changing if the
 * pin is high.  The interrupt pins are available on the parallel port
 * or their own header block on the top edge of the SR DSP32C boards.
 * 
 * Note that while the DSP32C automatically saves and restores the
 * entire state of the DAU when interrupting, the values of the CAU
 * registers are NOT saved and restored.  This means you can pass
 * parameters from the main line code to the isr in the rN registers.
 * 
 * See page 7-29 in the AT&T DSP32C Information manual for more
 * information about interrupts.
 * 
 *
 *
 */




/* PCW interrupt mask bits:
 *
 * These constants define the bits that must be set in the DSP32C's PCW
 * register to enable a particular interrupt.
 *
 * Be careful not to corrupt the PCW's other bits when changing values.
 *
 *
 */

constant

        ENABLE_INTREQ1          = 0x8000,
        ENABLE_PDF              = 0x4000,
        ENABLE_PDE              = 0x2000,
        ENABLE_IBF              = 0x1000,
        ENABLE_OBE              = 0x0800,
        ENABLE_INTREQ2          = 0x0400;






/* Main program:
 *
 * Set up and wait for interrupts.
 *
 *
 *
 */

main {

        register

                value   = r1,
                ivtp    = r22;  /* DSP32C dedicated ivtp pointer */



        /* Program the PCW register for zero wait states. */

        value = 0; pcw = value;



        /* Point ivtp to the interrupt vector table. */

        ivtp = IVTP_TABLE;



        /* Enable INTREQ1 and wait for interrupts. */

        value = pcw;

            value |= ENABLE_INTREQ1; pcw = value;


          wait:

                /* Put any foreground code in here. */


                goto wait;


}

long count = 0;







/* IVTP interrupt vector table:
 *
 * While you might want to program the ivtp table as a static array, it
 * is better to program it as a function.  That way you can emit goto's
 * directly into its table entries.
 * 
 * The routine name is just the starting address of the table and can be
 * treated as a constant to load into r22.
 *
 * If you like you can leave DSPASM's automatic nop generation on and
 * write the ivtp table without the explicit nops.
 *
 *
 *
 */

IVTP_TABLE {

    nops_off;

        goto isr_INTREQ1;               /* interrupt vector #1 */
        nop;

        goto isr_PDF;                   /* interrupt vector #2 */
        nop;

        goto isr_PDE;                   /* interrupt vector #3 */
        nop;

        goto isr_IBF;                   /* interrupt vector #4 */
        nop;

        goto isr_OBE;                   /* interrupt vector #5 */
        nop;

        goto isr_INTREQ2;               /* interrupt vector #6 */
        nop;

    nops_on;

}








/* Interrupt service routines:
 *
 * The code for the interrupt service routines is included here. Each
 * isr should end with an ireturn instruction.  This reenables the
 * interrupt, resuming mainline execution.
 *
 * Interrupt service routines not used here are simply stubbed out.
 *
 *
 */

isr_INTREQ1 {

        register c = r2;


        /* Increment the count. */

        c = *count; c += 1; *count = c;


        /* Return to the main line code. */

        ireturn;


}
isr_PDF {

        ireturn;

}
isr_PDE {

        ireturn;

}
isr_IBF {

        ireturn;

}
isr_OBE {

        ireturn;

}
isr_INTREQ2 {

        ireturn;

}
