/* FILE: dspmath.dsp            Copyright (c), Symmetric Research, 1998
 *
 * This is the DSP32C portion of the DSPMATH library.  See dspmath.c for
 * a general discussion on how the library works.
 *
 *
 * NOTES:
 *
 * * ASSEMBLE THIS WITH THE DSPASM /2 OPTION, FORWARD REFERENCING:
 *
 * This file has been written to use forward referencing and must be
 * assembled with the DSPASM /2 option.  
 *
 *
 * * ASSEMBLE THIS WITH THE DSPASM /h and /sc OPTIONS:
 *
 * The C part of the DSPMATH library requires a hexified version of the
 * object code for this file and also its symbol table in C include file
 * format.  Assembling with the DSPASM /h and /sc options generate both
 * of these files as dspmath.h and dspmath.dsc respectively.
 *
 *
 * * MAIN MUST BE AT LOCATION 0:
 *
 * At reset the DSP32C begins execution at location 0, and must have
 * code there for it to execute.  The DSPMATH library has been written
 * to expect main to be at location 0.  Don't change its location.
 *
 *
 * * INPUT_ARRAY MUST BE IN THE LOWER 16:
 *
 * Pointers and other small data items from the PC are passed to the
 * coprocessor board in the input array following main below.  Because
 * the direct addressing mode is used to reference the input data, it is
 * important that it be in the lower 16 bit address space.  Single word
 * instruction encoding limits the DSP32C direct addressing mode to 16
 * bits, and this file will not assemble correctly if the input array is
 * moved off to higher memory.
 *
 *
 * * LARGE DATA IS PASSED ON THE DATA STACK:
 *
 * Arrays and large data items are passed from the PC to DSP32C on the
 * data stack.  This is one large area of DSP32C memory for dumping large
 * arrays and data.   Pointers to items on the data stack are passed to
 * the DSP32C in the input_array in low memory.
 *
 * On the original 640K DSP32C board, the data stack should be programmed
 * to start at 0x800000, the start of memory partition B where a large
 * contiguous 512K byte block is located.
 *
 * On all of the new boards, DSP_400, DSP_MOD, and DSP_MUL, the data
 * stack should be programmed to follow the functions in low memory.
 * All of the new boards have a memory map with a single block of
 * memory starting at location 0.  They do not have the partition B
 * block starting at 0x800000.
 *
 * The constant DATA_STACK should be set to either 0x800000 or to
 * follow the functions below as appropriate for the board you are using.
 *
 *
 * * DEDICATED SUBROUTINE STACK AND RETURN ADDRESS REGISTERS:
 *
 * In addition to the PC data stack in high memory for large arrays, a
 * subroutine stack for calling helper functions local to this file is
 * also maintained in software.  This is required because the DSP32C
 * does not have dedicated stack pointer hardware for nested subroutine
 * calls and return addresses.
 *
 * The convention used here is that r13 is used as the stack pointer,
 * while r14 is used by DSPASM as the default return address register.
 * When a call is made, the DSP32C will place the return address into
 * r14, and on entry and exit to the called function, r14 should be
 * pushed and poped on the subroutine stack so subsequent calls can be
 * made without losing the return address.  Also, any other registers
 * that need to be saved and restored may be pushed and poped on the
 * stack.  Functions not making any calls do not need to save and
 * restore r14 on the subroutine stack.
 * 
 * NO other use should be made of r13 and r14, or else the thread of
 * execution may be lost.
 *
 *
 * * DO ONLY DOES 2048:
 *
 * In the functions below the do instruction is often used to achieve
 * the best perfomance.  Significant preformance improvements are
 * possible with do because instruction fetching is turned off while
 * looping.  However, the repeat count for do must be less than or equal
 * to 2048 = 0x800.  This limitation forces breaking up longer repeat
 * counts into blocks of 2048.  Functions like to_dspf are somewhat more
 * complicated than they would be if do could take a 24 bit repeat
 * count.  The code fragment from to_dspf is repeated often in the
 * other functions.
 *
 *
 *
 */



/* Assembly constants.
 *
 * These defined constants are used for readability further on down in the
 * code.
 *
 */

constant

        /* Memory usage:
         *
         * These constants are the base addresses of code and data segments
         * below.
         *
         */

         RESET          = 0x000000,             /* reset location at 0  */
         MEM_CONSTANTS  = 0xFFFA00,             /* in memory constants  */



        /* Status returns.
         *
         * These constants are returned in the PIR register with the
         * status of a calculation.  When an error occurs, a calculation
         * is halted and not carried to completion.
         *
         */

         SUCCESS        = 1,
         ERROR_INV      = 2,
         ERROR_SQRT     = 3,
         ERROR_LOG      = 4,


        /* DAUC register values.
         *
         * Setting the dauc to these values sets the conversion mode
         * on the int and int24 instructions.
         *
         */

         ROUND          = 0x00,
         TRUNCATE       = 0x10,


         ; /* end of assembly constants */





/* In memory constants and tables.
 *
 * These constants and tables are used generally by DSPMATH further on down
 * in the code.
 *
 */

<MEM_CONSTANTS>

dspf
        /* general constants */

        ZERO            =   0.000000000,
        ONE             =   1.000000000,
        TWO             =   2.000000000,
        THREE           =   3.000000000,
        FOUR            =   4.000000000,
        FIVE            =   5.000000000,
        SIX             =   6.000000000,
        HALF            =   0.500000000,
        MHALF           = - 0.500000000,
        THIRD           =   0.333333333,
        QUARTER         =   0.250000000,
        FIFTH           =   0.200000000,
        SIXTH           =   0.166666667,
        PI              =   3.141592654,
        TWOPI           =   6.283185307,
        HALFPI          =   1.570796327,


        /* special function coefficients */

        inv_coeff[] = {

                - 1.60138661,
                  0.82371354,
                  0.85221935,
                - 1.00000000,

                 },


        sqrt_coeff[] = {

                  1.000000000,
                  1.414213600,

                - 0.382891660,
                  1.187877720,
                  0.049693281,
                - 1.921554740,
                  2.066790900,
                  0.500000000,
                  1.500000100,
                },

                   sqrt_x, sqrt_temp[2],


        sin_coeff[] = {

                /* constants for angle reduction */

                  0.15915494,   /* INV_TWOPI    */
                  0.25000000,   /* QUARTER      */
                  6.28318531,   /* TWOPI        */
                  1.57079633,   /* HALFPI       */

                /* polynomial coefficients */

                  0.2601903036E-5,
                - 0.1980741820E-3,
                  0.8333025139E-2,
                - 0.1666665668E+0,

                },

        log10_coeff[] = {

                  0.301029995E+0,
                - 1.332183622E+0,
                  2.236407471E+1,
                  4.909769401E+0,
                - 1.205022337E+1,
                - 1.441930397E+1,
                - 2.896563148E+1,
                  2.542144201E+1,
                  4.772579384E+0,
                - 0.700519706E+0,

                },

                   log10_mantissa,
                   log10_exponent,


        /* FFT tables */

        forward_table[] = { /* n <= 16 point butterfly tables */

                 0.0000000000,  1.0000000000,  /* cos and sin of 2*PI*k/n */
                 0.7071067812,  0.7071067812,
                 0.0000000000,  1.0000000000,
                -0.7071067812,  0.7071067812,
                 0.9238795325,  0.3826834324,
                 0.7071067812,  0.7071067812,
                 0.3826834324,  0.9238795325,
                 0.0000000000,  1.0000000000,
                -0.3826834324,  0.9238795325,
                -0.7071067812,  0.7071067812,
                -0.9238795325,  0.3826834324,

                },

        inverse_table[] = {

                 0.0000000000, -1.0000000000,  /* cos and -sin of 2*PI*k/n */
                 0.7071067812, -0.7071067812,
                 0.0000000000, -1.0000000000,
                -0.7071067812, -0.7071067812,
                 0.9238795325, -0.3826834324,
                 0.7071067812, -0.7071067812,
                 0.3826834324, -0.9238795325,
                 0.0000000000, -1.0000000000,
                -0.3826834324, -0.9238795325,
                -0.7071067812, -0.7071067812,
                -0.9238795325, -0.3826834324,

                },

        forward_twids[] = { /* 32 <= n twiddle tables */

                 9.8078528040E-1,  1.9509032202E-1, /* cos, sin of 2*PI/n */
                 9.9518472667E-1,  9.8017140330E-2,
                 9.9879545621E-1,  4.9067674327E-2,
                 9.9969881870E-1,  2.4541228523E-2,
                 9.9992470184E-1,  1.2271538286E-2,
                 9.9998117528E-1,  6.1358846492E-3,
                 9.9999529381E-1,  3.0679567630E-3,
                 9.9999882345E-1,  1.5339801863E-3,
                 9.9999970586E-1,  7.6699031874E-4,
                 9.9999992647E-1,  3.8349518757E-4,
                 9.9999998162E-1,  1.9174759731E-4,
                 9.9999999540E-1,  9.5873799096E-5,
                 9.9999999885E-1,  4.7936899603E-5,

                 },

        inverse_twids[] = {

                 9.8078528040E-1, - 1.9509032202E-1, /* cos, -sin of 2*PI/n */
                 9.9518472667E-1, - 9.8017140330E-2,
                 9.9879545621E-1, - 4.9067674327E-2,
                 9.9969881870E-1, - 2.4541228523E-2,
                 9.9992470184E-1, - 1.2271538286E-2,
                 9.9998117528E-1, - 6.1358846492E-3,
                 9.9999529381E-1, - 3.0679567630E-3,
                 9.9999882345E-1, - 1.5339801863E-3,
                 9.9999970586E-1, - 7.6699031874E-4,
                 9.9999992647E-1, - 3.8349518757E-4,
                 9.9999998162E-1, - 1.9174759731E-4,
                 9.9999999540E-1, - 9.5873799096E-5,
                 9.9999999885E-1, - 4.7936899603E-5,

                 }

                 ; /* end of in memory constants */





/* Main entry point.
 *
 * Set up the subroutine stack, and call the function specified by the PC.
 *
 *
 */

<RESET>

main { /* <- THIS *MUST* BE THE RESET ENTRY POINT */

        struct input_array

                { long func /* , ... */ ; } input;

        register

                p       = r1,
                value   = r2,

                sp      = r13;



        /* Set the wait states.  See set_waits below for pcw_waits. */

        p = pcw_waits; value = *p; pcw = value;


        /* Point r13 to the subroutine stack, and call the function. */

        sp = stack;

                r1 = *input.func; call r1;


        /* Set the pir register, and wait for the PC. */

        r1 = SUCCESS; pir = r1; wait: goto wait;

}

long
        /* PC input structure with function selection and pointers. */

        input_array[50] = { 0 };





/* Helper functions:
 *
 * These functions do various jobs for the main library functions below.
 *
 */

to_dspf {

        /* Convert an ieee array to dspf floats. */

        register

                n       = r1,
                src     = r2,
                dest    = r3;


        for ( dest = src ; n > 0x800 ; n -= 0x800 )

                do 0x800 { *dest++ = a0 = dsp( *src++); }

        n--; do n+1

                *dest++ = a0 = dsp( *src++);

        return;

}
to_ieee {

        /* Convert a dpsf array to ieee floats. */

        register

                n       = r1,
                src     = r2,
                dest    = r3;


        for ( dest = src ; n > 0x800 ; n -= 0x800 )

                do 0x800 { *dest++ = a0 = ieee( *src++); }

        n--; do n+1

                *dest++ = a0 = ieee( *src++);

        return;

}
to_word {

        /* Convert a dspf array to 16 bit integers. */

        register

                n       = r1,
                src     = r2,
                dest    = r3;


        for ( dest = src ; n > 0x800 ; n -= 0x800 )

                do 0x800 { *dest++ = a0 = int( *src++); }

        n--; do n+1

                *dest++ = a0 = int( *src++);

        return;

}
inv {

        /* Compute a0 = 1/a0. */

        register

                c       = r1,
                sp      = r13;


        /* Save registers on the stack. */

        *sp++ = c;
        *sp++ = a1 = a1;
        *sp++ = a2 = a2;
        *sp   = a3 = a3;


        /* Test and trap on divide by zero. */

        a1 = a0; nop; nop; nop;

                if ( aeq )

                    { r1 = ERROR_INV; pir = r1; inv_error: goto inv_error; }


        /* Compute the inverse. */

        a1 = seed( a0); c = inv_coeff; nop;

        a2 = *c++ + a0*a1;
        a1 = a1 * *c++;
        nop;
        a2 = *c++ + a2*a2;
        a3 = a1 * a0;
        nop;
        a2 = a2 * a1;
        a1 = *c + a2*a3;
        nop;
        nop;
        a0 = a2 - a1*a2;


        /* Restore registers and return. */

        a3 = *sp--;
        a2 = *sp--;
        a1 = *sp--;
         c = *sp;

        return;

}
sqrt {

        /* Compute a0 = sqrt( a0). */

        register

                c       = r1,
                x       = r2,
                exp     = r3,
                iexp    = r4,
                temp    = r5,

                sp      = r13;


        /* Save registers on the stack. */

        *sp++ = c;
        *sp++ = x;
        *sp++ = exp;
        *sp++ = iexp;
        *sp++ = temp;
        *sp++ = a1 = a1;
        *sp++ = a2 = a2;
        *sp   = a3 = a3;


        /* Test and trap if a0 < 0. */

        a1 = a0; nop; nop; nop;

            if ( alt )

                { r1 = ERROR_SQRT; pir = r1; sqrt_error: goto sqrt_error; }


        /* Compute the square root with a Cheby approx. */

        x = sqrt_x;

                *x = a0 = a0;

        c = sqrt_coeff; temp = sqrt_temp;

                *temp++ = a1 = *c++;
                *temp-- = a1 = *c++;

        exp = *x.b;

                iexp = -exp; *temp = iexp.b;


        a3 = *x * *temp;

         exp >>= 1; if ( cs )

                { temp += 4; }


        a2 = a3 * a3;
        a1 = a3 * *c++;
        a1 = a1 + *c++;
        a1 = a1 + a2 * *c++;
        a2 = a3 * *c++;
        a2 = a2 + *c++;
        a1 = a2 + a2 * a1;
        a3 = a3 * *c++;
        exp += 0x40;
        a1 = a3 * a1;
        a2 = a1 * a1;
        *temp = exp.b;
        a1 = a1 * *temp;
        a2 = *c - a2 * a3;
        nop;
        nop;
        a0 = a1 * a2;


        /* Restore registers and return. */

          a3 = *sp--;
          a2 = *sp--;
          a1 = *sp--;
        temp = *sp--;
        iexp = *sp--;
         exp = *sp--;
           x = *sp--;
           c = *sp;

        return;

}
sin {

        /* Compute a0 = sin( a0), where a0 is in radians. */

        register

                c       = r1,
                sp      = r13;


        /* Save registers on the stack. */

        *sp++ = c;
        *sp++ = a1 = a1;
        *sp++ = a2 = a2;
        *sp   = a3 = a3;


sin_poly:

        /* Reduce a0 to an angle between [-PI/2,PI/2]. */

        c = sin_coeff;                  /* c -> coeff table     */

        a0 = *c++ + a0 * *c++;          /* x = 0.25 + x/(2PI)   */

        dauc = ROUND;                   /* set the int24 mode   */

        a1 = int24( a0);                /* x = x - int( x)      */
        a1 = float24( a1);
        a0 = -a1 + a0;

        a1 = - a0;                      /* x = abs( x)          */
        a0 = ifagt( a1);
        nop;
        nop;

        a0 = - *c++ + a0 * *c++;        /* x = x*2PI - PI/2     */
        nop;
        nop;


        /* Compute sin with a polynomial approx on [-PI/2,PI/2]. */

        a2 = a0*a0;
        nop;
        nop;
        a3 = a2*a0;
        a2 = *c++ + a2 * *c++;
        a1 = *c++ + a2 * *c++;
        a3 = a3*a0;
        a2 = a2*a3;
        a1 = a0 + a1*a3;
        nop;
        a0 = a1 + a2*a3;

            a0 = round( a0);


        /* Restore registers and return. */

        a3 = *sp--;
        a2 = *sp--;
        a1 = *sp--;
         c = *sp;

        return;

}
cos {

        /* Compute a0 = cos( a0), where a0 is in radians. */

        register

                c       = r1,
                sp      = r13;


        /* Save registers on the stack. */

        *sp++ = c;
        *sp++ = a1 = a1;
        *sp++ = a2 = a2;
        *sp   = a3 = a3;


        /* Actually compute sin( a0 + PI/2) = cos( a0). */

        c = HALFPI; a0 = *c + a0;

                        goto sin_poly;

}
log10 {

        /* Compute a0 = log10( a0), where a0 > 0. */

        register

                c       = r1,
                M       = r2,
                E       = r3,
                temp    = r4,

                sp      = r13;


        /* Save registers on the stack. */

        *sp++ = c;
        *sp++ = M;
        *sp++ = E;
        *sp++ = temp;
        *sp++ = a1 = a1;
        *sp++ = a2 = a2;
        *sp   = a3 = a3;


        /* Test and trap if a0 <= 0. */

        a1 = a0; nop; nop; nop;

            if ( ale )

                { r1 = ERROR_LOG; pir = r1; log_error: goto log_error; }



        /* Point to the coefficient table. */

        c = log10_coeff; c += 24; /* c -> c[8] */



        /* Compute the mantissa M/2. */

        M = log10_mantissa;

                *M = a0 = a0; nop; nop; temp = 7F; *M = temp.b;

                        a3 = *M;


        /* Compute the exponent E+1 as a float. */

        E = log10_exponent;

                *E = a0 = a0; nop; nop; nop; temp = *E.b;

                        temp -= 7F; *E = temp; *E = a0 = float24( *E);


        /* Compute the polynomial approximation. */

        a0 = a3*a3;
        a1 = *c-- + a3 * *c--;
        a2 = *c-- + a3 * *c--;
        a0 = a0 * a3;
        a1 = *c-- + a1 * a3;
        a3 = *c-- + a3 * *c--;
        a2 = *c-- + a2 * *M;
        a1 = a1 * a0;
        a3 = *c-- + a3 * *M;
        a0 = *E * *c;
        a2 = a3 + a2 * a0;
        a1 = a2 + a1 * a0;


        /* Add c9*(E-1) and store. */
        
        a0 = a1 + a0;


        /* Restore registers and return. */

          a3 = *sp--;
          a2 = *sp--;
          a1 = *sp--;
        temp = *sp--;
           E = *sp--;
           M = *sp--;
           c = *sp;

        return;

}





/* DSPMATH LIBRARY FUNCTIONS:
 *
 * These are the main functions called by the DSPMATH library.
 *
 *
 */

fft {
        /* Compute an FFT with dpsf floats as input and output. */

        struct input_array

                { long func, A, n, isign; } input;

        register

                N               = r1,
                A               = r2,

                A1              = r3,
                A2              = r4,
                Z1              = r5,
                Z2              = r6,

                i               = r7,
                j               = r8,

                t               = r9,

                zero            = r10,
                one             = r11,
                two             = r12,

                sp              = r13,          /* stack pointer        */
                ra              = r14,          /* return address       */

                n               = r15,
                k               = r16,
                p               = r17,

                inc             = r18,
                nover2          = r19,

                pp              = r20,
                table           = r21;


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Compute the in place bit reversal. */

        N = *input.n;
        A = *input.A;

        A1 = A;
        A2 = A;

        N *= 8;

        for ( i = 0 , j = 0 , one = N/2 ; i < N ; i += 8 ) {

                /* Swap A[i] and A[j] if not done yet. */

                if ( j > i ) { Z1 = A1;

                        *Z1++ = a0 = *A2++;
                        *Z1-- = a0 = *A2--;
                        *A2++ = a0 = *A1++;
                        *A2-- = a0 = *A1--;

                        }

                /* Reverse inc j, and update pointers. */

                j #= one; A1 += 8; A2 = A+j;

                }


        /* Compute the FFT butterflys. */

        N = *input.n;

        zero = ZERO;
         one = ONE;
         two = TWO;

         N *= 8;


        /* Do all butterflys up to 16 points. */
                
        A = *input.A;

            i = *input.isign; /* point to forward or inverse tables */

              if ( i > 0 ) table = forward_table; else table = inverse_table;


        for ( pp = 0 ; pp < N ; pp += 0x80 /* = 16*8 */ ) {

                t = table;

                for ( n = 0x10 /* = 2*8 */ , p = 8 ; n <= 0x80 ; n *= 2 ) {

                        nover2 = n/2; inc = n-4;


                    /* special case k = 0 */

                        A1 = A; A2 = A1+nover2; Z1 = A1; Z2 = A2;

                        i = p-1; do i+1 {

                                *Z1++    = a0 = *A1++    + *A2++;
                                *Z1++inc = a0 = *A1--    + *A2--;
                                *Z2++    = a0 = *A1++    - *A2++;
                                *Z2++inc = a0 = *A1++inc - *A2++inc;

                                }

                    /* remaining values of k */

                        for ( k = 8 /* = 1*8 */ ; k < nover2 ; k += 8 ) {

                                a2 = *t++;
                                a3 = *t++;

                                A1 = A+k; A2 = A1+nover2; Z1 = A1; Z2 = A2;

                                i = p-1; do i+1 {

                                               a0 = *A1++ +   a2 * *A2++;
                                    *Z1++    = a0 =    a0 -   a3 * *A2--;
                                               a1 = *A1-- +   a3 * *A2++;
                                    *Z2++    = a0 =  - a0 + *two * *A1++;
                                    *Z1++inc = a1 =    a1 +   a2 * *A2++inc;
                                    *Z2++inc = a1 =  - a1 + *two * *A1++inc;

                                    }
                                }
                        p /= 2;

                        }

                A += 0x80;

                }


        /* Compute the remaining butterflys, n = 32, 64, .... */

            i = *input.isign; /* point to forward or inverse twids */

                if ( i > 0 ) t = forward_twids; else t = inverse_twids;


        for ( n = 0x100 /* = 32*8 */ , p = N/0x100 ; n <= N ; n *= 2 ) {

            A = *input.A; nover2 = n/2; inc = n-4;

                a2 = *one;
                a3 = *zero;

                for ( k = 0 ; k < nover2 ; k += 8 ) {

                        A1 = A+k; A2 = A1+nover2; Z1 = A1; Z2 = A2;

                        i = p-1; do i+1 {

                                           a0 = *A1++ +   a2 * *A2++;
                                *Z1++    = a0 =    a0 -   a3 * *A2--;
                                           a1 = *A1-- +   a3 * *A2++;
                                *Z2++    = a0 =  - a0 + *two * *A1++;
                                *Z1++inc = a1 =    a1 +   a2 * *A2++inc;
                                *Z2++inc = a1 =  - a1 + *two * *A1++inc;

                                }

                        /* Twiddle cexp = (a2,a3). */

                        a0 = a2 * *t++; a2 = a0 - a3 * *t;
                        a0 = a2 * *t--; a3 = a0 + a3 * *t;

                        }

                /* Point to the next twid and loop to next n. */

                t += 8; p /= 2;

                }


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
rfft {

        /* Compute a real FFT with dpsf floats as input and output. */

        struct input_array

                { long func, x, n, isign; } input;

        register

                n               = r1,
                i               = r2,
                t               = r3,

                A               = r4,
                A1              = r5,
                A2              = r6,
                Z1              = r7,
                Z2              = r8,

                c1              = r9,
                c2              = r10,
                two             = r11,
                isign           = r12,

                sp              = r13,          /* stack pointer        */
                ra              = r14,          /* return address       */

                inc             = r15,
                nover4          = r16;


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert *input.n to the input viewed as complex. */

        n = *input.n; n /= 2; *input.n = n;



        /* Set up for a forward or inverse real FFT. */

        isign = *input.isign;

            if ( isign >= 0 ) {

                    call fft; /* NOTE: most regs are trashed by fft */

                        t = forward_twids; c2 = HALF; }
            else
                    { t = inverse_twids; c2 = MHALF; }


        c1 = HALF;

        n = *input.n;

        for ( i = 0x10 ; i < n ; i *= 2 ) /* point to the twid */

                t += 8;



        /* Process the input. */

        n = *input.n;
        A = *input.x;

            n *= 8; A1 = A+8; A2 = A+n; A2 -= 8; inc = 0xC; inc = -inc;

                Z1 = A1; Z2 = A2;

                    two = TWO; a2 = *t++; a3 = *t--; /* (a2,a3) = cexp */


        for ( i = 0x10 ; i < n ; i += 0x10 ) {


                /* Untangle. */

                *Z1++ = a0 =   *A1++ + *A2++;
                *Z1-- = a0 =   *A1   - *A2;
                *Z2++ = a0 =   *A1-- + *A2--;
                *Z2-- = a0 = - *A1   + *A2;

                    nop; nop;

                 *Z1++ = a0 = *c1 * *A1++;
                 *Z1-- = a0 = *c1 * *A1--;
                 *Z2++ = a0 = *c2 * *A2++;
                 *Z2-- = a0 = *c2 * *A2--;

                    nop; nop;


                /* Compute the butterfly. */

                           a0 = *A1++ +   a2 * *A2++;
                *Z1++    = a0 =    a0 -   a3 * *A2--;
                           a1 = *A1-- +   a3 * *A2++;
                *Z2++    = a0 =  - a0 + *two * *A1++;
                *Z1++    = a1 =    a1 +   a2 * *A2++inc;
                *Z2++inc = a1 =    a1 - *two * *A1++;


                /* Twiddle cexp = (a2,a3). */

                a0 = a2 * *t++; a2 = a0 - a3 * *t;
                a0 = a2 * *t--; a3 = a0 + a3 * *t;

                }


        /* Finish off the transform. */

        A1 = A; A2 = A+4; Z1 = A;

            *Z1++ = a0 = *A1 + *A2;
            *Z1-- = a0 = *A1 - *A2;


        isign = *input.isign; if ( isign < 0 ) { /* then inverse */

                *Z1++ = a0 = *c1 * *A1;
                *Z1-- = a0 = *c1 * *A2;

                /* Take the inverse complex transform. */

                call fft;

                }


        /* Restore *input.n back to the input view as reals. */

        n = *input.n; n *= 2; *input.n = n;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
fourier {


        /* Compute an fft, with ieee floats as input and output. */

        struct input_array

                { long func, A, n, isign; } input;

        register

                sp      = r13,          /* stack pointer        */
                ra      = r14;          /* return address       */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert A to dspf floats. */

        r1 = *input.n; r1 *= 2; r2 = *input.A; call to_dspf;


        /* Compute the transform. */

        call fft;


        /* Convert A back to ieee floats. */

        r1 = *input.n; r1 *= 2; r2 = *input.A; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
rfourier {

        /* Compute the real FFT of an input array of ieee floats. */

        struct input_array

                { long func, x, n, isign; } input;

        register

                sp      = r13,          /* stack pointer        */
                ra      = r14;          /* return address       */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;


        /* Compute the transform. */

        call rfft;


        /* Convert the output back to ieee floats. */

        r1 = *input.n; r2 = *input.x; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
spectrum {

        /* Compute the spectrum of a real array. */

        struct input_array

                { long func, x, n, isign, taper; } input;

        register

                n               = r1,
                i               = r2,
                x               = r3,
                A               = r4,
                Adest           = r5,

                iend            = r6,
                taper           = r7,

                zero            = r8,
                four            = r9,

                nover2          = r10,

                sp              = r13,          /* stack pointer        */
                ra              = r14;          /* return address       */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Set the taper flag/pointer. */

        taper = *input.taper;


        /* Convert the input array to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;

        if ( taper != 0 )

                { r1 = *input.n; r2 = *input.taper; call to_dspf; }



        /* Taper with either a user specified or linear ramp taper. */

        if ( taper != 0 ) {

                n = *input.n;
                x = *input.x;

                for ( i = 0 ; i < n ; i++ )

                        *x++ = a0 = *taper++ * *x;
                }

        else {

                /* Compute the slope for a linear taper in a1. */

                zero = ZERO; four = FOUR;

                n = input.n;

                    a0 = float24( *n); call inv; a1 = *four * a0;


                /* Taper the first n/4 points. */

                 n = *input.n;
                 x = *input.x;
                a0 = *zero;

                i = n/4; while ( i-- > 0 )

                        { *x++ = a2 = *x * a0; a0 = a0 + a1; }


                /* Taper the last n/4 points. */

                 x = *input.x;
                a0 = *zero;

                        iend = n-1; iend *= 4; x += iend;

                i = n/4; while ( i-- > 0 )

                        { *x-- = a2 = *x * a0; a0 = a0 + a1; }

                }


        /* Compute the real FFT. */

        call rfft;


        /* Compute the complex absolute value. */

            n = *input.n;
            A = *input.x;
        Adest = *input.x;

         zero = ZERO;

                /* Get the packed DC value. */

                a0 = *A; a1 = - *A++; *Adest++ = a0 = ifagt( a1);


                /* Save the packed Nyquist value in a1. */

                a1 = *A; a2 = - *A++; a1 = ifagt( a2);


                /* Compute the complex abs value up to the NYQUIST. */

                for ( i = 1 , nover2 = n/2 ; i < nover2 ; i++ ) {

                        a0 = *A * *A; A += 4; nop; a0 = a0 + *A * *A;

                                call sqrt; *Adest++ = a0 = a0;

                          A +=  4;
                        }

                /* Output the Nyquist point. */

                *Adest++ = a1 = a1;


                /* Zero off the remaining points. */

                for ( i = nover2+1 ; i < n ; i++ )

                        *Adest++ = a0 = *zero;



        /* Convert the output back to ieee floats. */

        r1 = *input.n; r2 = *input.x; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
convolve {

        /* Convolve two real time series with a filter using FFT's. */

        struct input_array

                { long func, A, n, isign, x1, x2, filter, fflag; } input;

        register

                i               = r1,
                n               = r2,
                f               = r3,
                A               = r4,
                Z               = r5,
                x1              = r6,
                x2              = r7,
                fflag           = r8,
                isign           = r9,
                nover2          = r10,

                sp              = r13,          /* stack pointer        */
                ra              = r14;          /* return address       */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x1, x2, and the filter to dspf floats. */

        r1 = *input.n; r2 = *input.x1; call to_dspf;
        r1 = *input.n; r2 = *input.x2; call to_dspf;

        fflag = *input.fflag;

            if ( fflag == 1 )

                { r1 = *input.n; r2 = *input.filter; call to_dspf; }



        /* Pack the two real arrays x1 and x2 into a single complex in A. */

         A = *input.A;
        x1 = *input.x1;
        x2 = *input.x2;

         n = *input.n; while ( n-- > 0 )

                { *A++ = a0 = *x1++; *A++ = a0 = *x2++; }



        /* Transform A. */

        isign = 1; *input.isign = isign; call fft;



        /* Multiply by the packed complex filter. */

                A = *input.A;
                Z = *input.A;
                f = *input.filter;
        
                    n = *input.n; nover2 = n/2;


                /* Special case i = 0, and save Nyquist value in a1. */

                *Z++ = a0 = *f   * *A++;
                *Z++ = a0 = *f++ * *A++;

                        a1 = *f++;


                /* Multiply up to the Nyquist, i = n/2. */

                for ( i = 1 ; i < nover2 ; i++ ) {

                        a0 = *f++ * *A++; *Z++ = a0 = a0 - *f-- * *A;
                        a0 = *f++ * *A--; *Z++ = a0 = a0 + *f-- * *A;

                            f += 8; A += 8;
                        }
                        

                /* Special case the Nyquist point, i = n/2. */

                *Z++ = a0 = a1 * *A++;
                *Z++ = a0 = a1 * *A++;


                /* Multiply the rest, moving f backward and conjugating it. */

                f -= 8;

                for ( i = nover2+1 ; i < n ; i++ ) {

                        a0 = *f++ * *A++; *Z++ = a0 = a0 + *f-- * *A;
                        a0 = *f++ * *A--; *Z++ = a0 = a0 - *f-- * *A;

                            f -= 8; A += 8;
                        }


        /* Transform back. */

        isign = 1; isign = - isign; *input.isign = isign; call fft;


        /* Unpack the two real arrays from the complex array A. */

         A = *input.A;
        x1 = *input.x1;
        x2 = *input.x2;

         n = *input.n;

            while ( n-- > 0 )

                { *x1++ = a0 = *A++; *x2++ = a0 = *A++; }


        /* Convert the output back to ieee floats. */

        r1 = *input.n; r2 = *input.x1; call to_ieee;
        r1 = *input.n; r2 = *input.x2; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
add {
        /* Add two vectors together returning the result. */

        struct input_array

                { long func, n, x, y, z; } input;

        register

                n       = r1,
                x       = r2,
                y       = r3,
                z       = r4,

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x and y to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;
        r1 = *input.n; r2 = *input.y; call to_dspf;


        /* Add the two vectors together. */

        x = *input.x;
        y = *input.y;
        z = *input.z;

                for ( n = *input.n ; n > 0x800 ; n -= 0x800 )

                        do 0x800 { *z++ = a0 = *x++ + *y++; }

                n--; do n+1

                        *z++ = a0 = *x++ + *y++;


        /* Convert z back to ieee floats. */

        r1 = *input.n; r2 = *input.z; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
sub {

        /* Add two vectors together returning the result. */

        struct input_array

                { long func, n, x, y, z; } input;

        register

                n       = r1,
                x       = r2,
                y       = r3,
                z       = r4,

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x and y to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;
        r1 = *input.n; r2 = *input.y; call to_dspf;



        /* Add the two vectors together. */

        x = *input.x;
        y = *input.y;
        z = *input.z;

                for ( n = *input.n ; n > 0x800 ; n-= 0x800 )

                        do 0x800 { *z++ = a0 = *x++ - *y++; }

                n--; do n+1

                        *z++ = a0 = *x++ - *y++;



        /* Convert z back to ieee floats. */

        r1 = *input.n; r2 = *input.z; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
xpcy {

        /* Compute x + c*y, where x and y are vectors and c a float. */

        struct input_array

                { long func, x, c, y, n, z; } input;

        register

                n       = r1,
                x       = r2,
                c       = r3,
                y       = r4,
                z       = r5,

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x and y to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;
        r1 = *input.n; r2 = *input.y; call to_dspf;


        /* Compute z = x + c*y. */

        x = *input.x;
        y = *input.y;
        z = *input.z;

        c = input.c; a0 = dsp( *c);

                for ( n = *input.n ; n > 0x800 ; n -= 0x800 )

                        do 0x800 { *z++ = a1 = *x++ + a0 * *y++; }

                n--; do n+1

                        *z++ = a1 = *x++ + a0 * *y++;



        /* Convert z back to ieee floats. */

        r1 = *input.n; r2 = *input.z; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
scale {

        /* Scale a vector x by c, returning the specified format. */

        struct input_array

                { long func, n, x; dspf c; long format; } input;

        register

                n       = r1,
                x       = r2,
                xdest   = r3,
                c       = r4,
                format  = r5,

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;



        /* Scale x. */

        x = *input.x;

          xdest = x;

          c = input.c; a0 = dsp( *c);

                for ( n = *input.n ; n > 0x800 ; n -= 0x800 )

                        do 0x800 { *xdest++ = a1 = a0 * *x++; }

                n--; do n+1

                        *xdest++ = a1 = a0 * *x++;



        /* Convert x back to the requested output format. */

        format = *input.format;

                r1 = *input.n; r2 = *input.x; r3 = *input.x;


        if ( format == 0 )

                { dauc = TRUNCATE; call to_word; }

        else if ( format == 1 )

                { dauc = ROUND; call to_word; }

        else

                call to_ieee; /* convert to ieee floats */



        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
inner {

        /* Compute the inner product of x and y. */

        struct input_array

                { long func, n, x, y, r; } input;

        register

                n       = r1,
                x       = r2,
                y       = r3,
                zero    = r4,
                result  = r5,

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x and y to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;
        r1 = *input.n; r2 = *input.y; call to_dspf;


        /* Compute the inner product. */

        x = *input.x;
        y = *input.y;

            zero = ZERO; a0 = *zero;


                for ( n = *input.n ; n > 0x800 ; n -= 0x800 )

                        do 0x800 { a0 = a0 + *x++ * *y++; }

                n--; do n+1

                        a0 = a0 + *x++ * *y++;

                                nop;


        /* Return the result as an ieee float. */

        result = *input.r; *result = a0 = ieee( a0);


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
vmult {

        /* Multiply x and y, entry by entry. */

        struct input_array

                { long func, n, x, y, z; } input;

        register

                n       = r1,
                x       = r2,
                y       = r3,
                z       = r4,

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x and y to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;
        r1 = *input.n; r2 = *input.y; call to_dspf;


        /* Multiply x and y, entry by entry, leaving the result in z. */

        x = *input.x;
        y = *input.y;
        z = *input.z;

                for ( n = *input.n ; n > 0x800 ; n -= 0x800 )

                        do 0x800 { *z++ = a0 = *x++ * *y++; }

                n--; do n+1

                        *z++ = a0 = *x++ * *y++;


        /* Convert z back to ieee floats. */

        r1 = *input.n; r2 = *input.z; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
mmult {

        struct input_array

                { long func, A, x, y, n, m, n2, nm; } input;

        register

                n       = r1,
                m       = r2,
                A       = r3,
                x       = r4,
                y       = r5,

                zero    = r6,

                i       = r7,
                j       = r8,
                k       = r9,

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert the input to dspf floats. */

        r1 = *input.n2; r2 = *input.A; call to_dspf;
        r1 = *input.nm; r2 = *input.x; call to_dspf;


        /* Compute Ax for the m x vectors. */

        n = *input.n;
        m = *input.m;
        A = *input.A;
        y = *input.y;

        zero = ZERO;

        for ( k = 0 ; k < m ; k++ ) {

                for ( i = 0 ; i < n ; i++ ) {

                        a0 = *zero; x = *input.x; j = n-1;

                                do j+1 a0 = a0 + *A++ * *x++;

                        *y++ = a0 = a0;
                        }

                *input.x = x; A = *input.A;

                }


        /* Convert the output back to ieee floats. */

        r1 = *input.nm; r2 = *input.y; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
AtranA {

        struct input_array
                
                { long func, A, b, ATA, ATb, n, m, nm, m2; } input;
        
        register

                i       = r1,
                j       = r2,
                k       = r3,
                A       = r4,
                AT      = r5,
                ATA     = r6,
                b       = r7,
                ATb     = r8,
                Acol    = r9,
                ATcol   = r10,

                zero    = r11,

                sp      = r13,                  /* stack pointer        */
                ra      = r14,                  /* return address reg   */

                n       = r15,
                m       = r16,
                mbytes  = r17;


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert the input to dspf floats. */

        r1 = *input.nm; r2 = *input.A; call to_dspf;
        r1 = *input.n;  r2 = *input.b; call to_dspf;



        /* Compute A transpose times A, ATA. */

             n = *input.n;
             m = *input.m;

          zero = ZERO;
        mbytes = m*4;

         ATcol = *input.A;
           ATA = *input.ATA;

        for ( i = 0 ; i < m ; i++ ) {

                Acol = *input.A;

                for ( j = 0 ; j < m ; j++ ) {

                        AT = ATcol; A = Acol; a0 = *zero;

                                k = n-1; do k+1

                                        a0 = a0 + *AT++mbytes * *A++mbytes;

                          *ATA++ = a0 = a0; Acol += 4;
                        }

                  ATcol += 4;
                }


        /* Compute A transpose times b, ATb. */

        Acol = *input.A;
         ATb = *input.ATb;

        for ( i = 0 ; i < m ; i++ ) {

                A = Acol; b = *input.b; a0 = *zero;

                        j = n-1; do j+1

                                a0 = a0 + *A++mbytes * *b++;

                  *ATb++ = a0 = a0; Acol += 4;
                }



        /* Convert the output back to ieee floats. */

        r1 = *input.m2; r2 = *input.ATA; call to_ieee;
        r1 = *input.m;  r2 = *input.ATb; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
lu_decompose {

        /* LU decompose a matrix with row operations. */

        struct input_array

                { long func, n, n2, A, b, p; } input;

        register

                i               = r1,
                ii              = r2,
                ipiv            = r3,
                j               = r4,
                pivot           = r5,
                arow            = r6,
                a               = r7,
                l               = r8,
                apiv            = r9,
                p               = r10,
                zero            = r11,

                sp              = r13,          /* stack pointer        */
                ra              = r14,          /* return address reg   */

                n               = r15,
                nm1             = r16,
                nbytes          = r17,
                ibytes          = r18;


        /* Save the return address. */

        *sp++ = ra;


        /* Compute the lu decomposition, with optional row pivoting. */

        n = *input.n; nm1 = n-1; nbytes = n*4;

         zero = ZERO;

         arow = *input.A;
        pivot = *input.p;

        for ( i = 0 ; i < nm1 ; i++ ) { ibytes = i*4;


                /* If pivoting, swap the maximum row into place. */

                if ( pivot != 0 ) {

                        /* Scan down the column for the pivot. */
                        
                        a = arow; apiv = arow; a0 = *zero;

                        for ( ii = i ; ii < n ; ii++ ) {

                                p = a+ibytes; a1 = - *p; a1 = ifalt( *p);
                                nop;
                                nop;

                                a2 = a1-a0; nop; nop; nop; if ( age )

                                        { a0 = a1; ipiv = ii; apiv = a; }

                                a += nbytes;
                                }

                        /* Swap the current row with the pivot. */
                        
                        a = arow;

                        for ( j = 0 ; j < n ; j++ )
                                { *a = a0 = *apiv; *apiv++ = a0 = *a++; }


                        /* Record the pivot. */
                        
                        *pivot++ = ipiv;

                        }

                /* Compute the decomposition with row operations. */

                a = arow+ibytes;
                l = a+nbytes;

                a0 = *a; call inv;
                
                for ( ii = i+1 ; ii < n ; ii++ ) {

                        *l++ = a1 = *l * a0; a += 4;

                        for ( j = i+1 ; j < n ; j++ )
                                *l++ = a2 = *l - a1 * *a++;

                        a = arow+ibytes;
                        l += ibytes;
                        }

                arow += nbytes;
                }


        /* If pivoting, clean up the last entry of pivot[]. */
        
        if ( pivot != 0 )

                { n--; *pivot = n; }


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
lu_solve {

        /* Solve an LU system with with forward and back substitution. */

        struct input_array

                { long func, n, n2, A, b, p; } input;

        register

                i               = r1,
                j               = r2,
                pivot           = r3,
                a               = r4,
                b               = r5,
                bpiv            = r6,
                arow            = r7,
                bend            = r8,

                zero            = r9,

                sp              = r13,          /* stack pointer        */
                ra              = r14,          /* return address reg   */

                n               = r15,
                nbytes          = r16;


        /* Save the return address. */

        *sp++ = ra;


        /* Compute the forward substitutions. */

            n = *input.n; nbytes = n*4; zero = ZERO;

         arow = *input.A;
            b = *input.b;
        pivot = *input.p;


        for ( i = 0 ; i < n ; i++ ) {
                
                if ( pivot != 0 ) {

                        bpiv = *input.b; j = *pivot++; j = j*4; bpiv += j;

                                { *b = a0 = *bpiv; *bpiv = a0 = *b; }
                        }


                b = *input.b; a = arow; a0 = *zero;

                /* for ( j = 0 ; j < i ; j++ ) */  j = i-1; if ( ge ) do j+1

                        { a0 = a0 + *a++ * *b++; } nop; nop;

                *b++ = a0 = *b - a0;

                arow += nbytes;
                }


        /* Compute the backward substitutions. */

        arow -= 4;
        bend = b-4;

        for ( i = 0 ; i < n ; i++ ) {

                a = arow;
                b = bend;

                a0 = *zero;

                /* for ( j = 0 ; j < i ; j++ ) */ j = i-1; if ( ge ) do j+1

                        { a0 = a0 + *a-- * *b--; } nop; nop;
                

                a1 = *b - a0; a0 = *a; call inv; *b = a0 = a1*a0;

                arow -= nbytes;
                }


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
gauss {

        /* LU matrix solver *WITH* optional pivoting. */

        struct input_array

                { long func, n, n2, A, b, p; } input;

        register

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert the input to dpsf floats. */

        r1 = *input.n2; r2 = *input.A; call to_dspf;
        r1 = *input.n;  r2 = *input.b; call to_dspf;


        /* Call the lu decomposer and solver above. */

        call lu_decompose;

                /**** Call solver only if b ???? *****/  call lu_solve;


        /* Convert the output back to ieee floats. */

        r1 = *input.n2; r2 = *input.A; call to_ieee;
        r1 = *input.n;  r2 = *input.b; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
tridiag {

        /* Tridiagonal matrix solver, LU decomposition *WITHOUT* pivoting. */

        struct input_array

                { long func, n, u, d, l, b; } input;

        register

                n       = r1,
                u       = r2,
                d       = r3,
                l       = r4,
                b       = r5,

                i       = r6,

                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert the input to dspf floats. */

        r1 = *input.n; r2 = *input.u; call to_dspf;
        r1 = *input.n; r2 = *input.d; call to_dspf;
        r1 = *input.n; r2 = *input.l; call to_dspf;
        r1 = *input.n; r2 = *input.b; call to_dspf;


        /* Compute the LU decomposition and forward substitutions. */

        n = *input.n;

        u = *input.u;                           /* u = &u[0]    */
        d = *input.d;                           /* d = &d[0]    */
        l = *input.l; l += 4;                   /* l = &l[1]    */
        b = *input.b;                           /* b = &b[0]    */

        a0 = *d++; call inv;

        *u++ = a1 = *u * a0;
        *b++ = a2 = *b * a0;

        for ( i = 1 ; i < n ; i++ ) {

                *d++ = a0 = *d - a1 * *l;

                call inv;

                *u++ = a1 = *u * a0;

                            a2 = *b - a2 * *l++; nop; nop;

                *b++ = a2 = a2 * a0;
                }


        /* Compute the back substitutions. */

        b -= 4;
        u -= 8;

        *b-- = a0 = *b;

        /* for ( i = n-2 ; i >= 0 ; i-- ) */ i = n-2; do i+1

                { *b-- = a0 = *b - a0 * *u--; nop; nop; }



        /* Convert the output back to ieee floats. */

        r1 = *input.n; r2 = *input.u; call to_ieee;
        r1 = *input.n; r2 = *input.d; call to_ieee;
        r1 = *input.n; r2 = *input.l; call to_ieee;
        r1 = *input.n; r2 = *input.b; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
fir {

        /* Compute the time domain FIR covolution. */

        struct input_array

                { long func, x, n, h, nfilter, y; } input;

        register

                n               = r1,
                i               = r2,
                j               = r3,
                x               = r4,
                h               = r5,
                y               = r6,
                xstart          = r7,
                nfilter         = r8,

                sp              = r13,          /* stack pointer        */
                ra              = r14;          /* return address reg   */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert the input x and h to dspf floats. */

        r1 = *input.n; r2 = *input.nfilter; r1 += r2;

                r2 = *input.x; call to_dspf; /* convert the leader too ! */


        r1 = *input.nfilter; r2 = *input.h; call to_dspf;




        /* Convolve h with x. */

              n = *input.n;
        nfilter = *input.nfilter;
              y = *input.y;

         xstart = *input.x;

                i = nfilter*4; xstart += i;


        for ( i = 0 ; i < n ; i++ ) {

                h = *input.h; x = xstart;


                a0 = *h++ * *x--;

                    j = nfilter-2; do j+1 a0 = a0 + *h++ * *x--;


                *y++ = a0 = a0; xstart += 4;

                }



        /* Convert the output back to ieee floats. */

        r1 = *input.n; r2 = *input.y; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
poly_array {

        /* Compute a polynomial at n points. */

        struct input_array

                { long func, nc, c, n, y, xstart, delta; } input;

        register

                n               = r1,
                nc              = r2,
                c               = r3,
                i               = r4,
                y               = r5,
                xstart          = r6,
                delta           = r7,

                sp              = r13,          /* stack pointer        */
                ra              = r14;          /* return address       */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert the input coefficients c[] to dspf floats. */

        r1 = *input.nc; r2 = *input.c; call to_dspf;


        /* Evaluate the polynomial at n points. */

            n = *input.n;
           nc = *input.nc;

            c = *input.c;       /* point to last entry of c[]   */

                nc--; i = nc*4; c += i; *input.c = c;

            y = *input.y;

        xstart = input.xstart;  a0 = dsp( *xstart);
         delta = input.delta;   a1 = dsp( *delta);

           nc--; /* decrement nc again for do loop count */


        for ( i = 0 ; i < n ; i++ ) {

                /* Compute a2 = y at x = a0 by Horner's rule. */

                c = *input.c; a2 = *c--; nop;

                        do nc+1 { a2 = *c-- + a0 * a2; nop; nop; }


                /* Save y and inc to next x. */

                *y++ = a2 = a2; a0 = a0 + a1;

                }


        /* Convert y to ieee floats. */

        r1 = *input.n; r2 = *input.y; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
varray {

        /* Evaluate a function at n equally spaced points. */

        struct input_array

                { long func, entry, n, y, xstart, delta; } input;

        register

                n               = r1,
                y               = r2,
                i               = r3,
                entry           = r4,
                xstart          = r5,
                delta           = r6,

                sp              = r13,          /* stack pointer        */
                ra              = r14;          /* return address       */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Go into a loop incrementing by delta. */

        entry = *input.entry;
            n = *input.n;
            y = *input.y;

        xstart = input.xstart;  a1 = dsp( *xstart);
         delta = input.delta;   a2 = dsp( *delta);

        for ( i = 0 ; i < n ; i++ )

                { a0 = a1; call entry; *y++ = a0 = a0; a1 = a1 + a2; }


        /* Convert y to ieee floats. */

        r1 = *input.n; r2 = *input.y; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
vfunc {

        /* Evaluate a function on each element of x[]. */

        struct input_array

                { long func, entry, x, n; } input;

        register

                n               = r1,
                x               = r2,
                entry           = r3,
                i               = r4,

                sp              = r13,          /* stack pointer        */
                ra              = r14;          /* return address       */


        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;


        /* Go into a loop computing f(x[]). */

            n = *input.n;
            x = *input.x;
        entry = *input.entry;

        for ( i = 0 ; i < n ; i++ )

                { a0 = *x; call entry; *x++ = a0 = a0; } 



        /* Convert x back to ieee floats. */

        r1 = *input.n; r2 = *input.x; call to_ieee;


        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
polar_to_xy {


        /* Convert polar coordinates to cartesian. */

        struct input_array

                { long func, rtheta, n, xy; } input;


        register

                i       = r1,
                n       = r2,
                rtheta  = r3,
                xy      = r4,
                
                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert the input to dspf floats. */

        n = *input.n;

                r1 = n*2; r2 = *input.rtheta; call to_dspf;


        /* Convert the polar coordinates to cartesian. */

             n = *input.n;
        rtheta = *input.rtheta;
            xy = *input.xy;

        for ( i = 0 ; i < n ; i++ ) {

                a2 = *rtheta++;         /* a2 = r       */
                a3 = *rtheta++;         /* a3 = theta   */

                        a0 = a3; call cos; *xy++ = a1 = a2*a0; 
                        a0 = a3; call sin; *xy++ = a1 = a2*a0;

                }


        /* Convert the output back to ieee floats. */

        n = *input.n;

                r1 = n*2; r2 = *input.xy; call to_ieee;



        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
interpolate {

        /* Interpolate an array. */

        struct input_array

                { long func, array1, n1, array2, n2, temp; } input;

        register

                i       = r1,
                j       = r2,
                n1      = r3,
                n2      = r4,
                temp    = r5,
                array1  = r6,
                array2  = r7,
                one     = r8,
                zero    = r9,
                
                sp      = r13,                  /* stack pointer        */
                ra      = r14;                  /* return address reg   */



        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert the input to dspf floats. */

        r1 = *input.n1; r2 = *input.array1; call to_dspf;



        /* Compute a0 = (n1-1)/(n2-1), the interpolation increment. */

        n2 = input.n2; a0 = float24( *n2); one = ONE;

            a0 = a0 - *one; call inv;

        n1 = input.n1; a1 = float24( *n1);

            a1 = a1 - *one; nop; nop; a0 = a1*a0; /* a0 = delta */



        /* Interpolate array1 onto array2. */

        array2 = *input.array2;
            n2 = *input.n2;

          temp = input.temp;
          zero = ZERO;
            a1 = *zero;

          dauc = TRUNCATE; /* set the int24 mode */


        for ( j = 0 ; j < n2 ; j++ ) {

                *temp = a2 = int24( a1); nop; nop; nop;

                        i = *temp; a2 = float24( *temp);


                    array1 = *input.array1; i *= 4; array1 += i;

                        a2 = a1 - a2; a3 = *array1-- - *array1++; nop; nop;


                *array2++ = a3 = *array1 + a2 * a3;

                    a1 = a1 + a0; /* x += delta */
                }




        /* Convert the output back to ieee floats. */

        r1 = *input.n2; r2 = *input.array2; call to_ieee;



        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}
spline {

        /* Compute the coefficients and interpolate a natural spline.
         *
         * The algorithm is taken from Applied Numerical Analysis by
         * C.F. Gerald and Numerical Recipies in C.
         *
         */

        struct input_array

            { long func, x, y, h, a, b, c, d, xx, yy, n, nn, xdiv; } input;


        register

                x       = r1,           /* input points         */
                y       = r2,
                h       = r3,           /* x intervals          */
                a       = r4,           /* spline coeff         */
                b       = r5,
                c       = r6,
                d       = r7,
                xx      = r8,           /* output points        */
                yy      = r9,
                xdiv    = r10,          /* interpolation inc    */

                six     = r11,
                two     = r11,
                zero    = r11,
                sixth   = r11,
                half    = r11,

                sp      = r13,          /* stack pointer        */
                ra      = r14,          /* return address reg   */

                i       = r15,
                n       = r16,          /* # of input points    */
                nn      = r17;          /* # of intrp points    */




        /* Save the return address on the stack. */

        *sp++ = ra;


        /* Convert x and y to dspf floats. */

        r1 = *input.n; r2 = *input.x; call to_dspf;
        r1 = *input.n; r2 = *input.y; call to_dspf;



        /* Compute h[i] = x[i+1] - x[i], the interval between points. */

        n = *input.n;
        x = *input.x;
        h = *input.h;

            i = n-2; do i+1 { *h++ = a0 = *x - *x++; }




        /* Compute
         *
         *   b[i+1] = 6*(y[i+2]-y[i+2])/h[i+1] - 6*(y[i+1]-y[i])/h[i]
         *
         * the right hand side of the tridiagonal spline system.  Save y
         * terms from iteration to iteration.
         *
         *
         */

        n = *input.n;
        b = *input.b; b += 4;
        y = *input.y;
        h = *input.h;


        a0 = *h++; call inv; six = SIX; a0 = a0 * *six;

            a2 = *y - *y++; nop; nop; a2 = a2*a0;


        n -= 2; for ( i = 0 ; i < n ; i++ ) {

                a0 = *h++; call inv; a0 = a0 * *six;

                a1 = a2;

                    a2 = *y - *y++; nop; nop; a2 = a2*a0; nop; nop;

                  *b++ = a0 = a2 - a1;
                }



        /* Setup the tridiagonal matrix to solve, ignoring the last
         * entries in a and c, where:
         *
         *    a[i] = h[i+1]   d[i] = 2*(h[i]+h[i+1])   c[i] = h[i+1]
         *
         * are the above diagonal, diagonal, and below diagonal entries.
         *
         */

          n = *input.n;
          h = *input.h; 
          a = *input.a;
          d = *input.d;
          c = *input.c;

        two = TWO;

        i = n-3; do i+1 {

                    a0 = *h + *h++;

                *a++ = a1 = *h;
                *c++ = a1 = *h;
                *d++ = a0 = a0 * *two;

                }


        /* Solve the tridiag system.  Compute first and last b below. */

        n = *input.n;

        a = *input.a;           /* lu decompose and forward subs */
        b = *input.b; b += 4;
        c = *input.c;
        d = *input.d;

        a0 = *d++; call inv;

        *a++ = a1 = *a * a0;
        *b++ = a2 = *b * a0;

        n -= 2; for ( i = 1 ; i < n ; i++ ) {

                *d++ = a0 = *d - a1 * *c;
                call inv;
                *a++ = a1 = *a * a0;
                a2 = *b - a2 * *c++; nop; nop;
                *b++ = a2 = a2 * a0;

                }


        b -= 4;                 /* back subs */
        a -= 8;

        *b-- = a0 = *b;

        n = *input.n;

            i = n-4; do i+1 { *b-- = a0 = *b - a0 * *a--; nop; nop; }



        /* Set first and last b values to zero. */

        n = *input.n;
        b = *input.b;

            zero = ZERO; i = n-1; i *= 4; 

                *b++i = a0 = *zero;
                *b    = a0 = *zero;




        /* Compute the spline abcd coeffcients using the second
         * derivatives s[i] now stored in b, where:
         *
         *  a[i] = (s[i+1]-s[i])/(6*h[i])
         *
         *  b[i] = s[i]/2
         *
         *  c[i] = (y[i+1]-y[i])/h[i] - (2*h[i]*s[i] + h[i]*s[i+1])/6
         *
         *  d[i] = y[i]
         *
         *
         */

        n = *input.n;  /* compute a and c */
        y = *input.y;
        h = *input.h;
        a = *input.a;
        b = *input.b;
        c = *input.c;

         sixth = SIXTH; a3 = *sixth; two = TWO;

        n--; for ( i = 0; i < n; i++ ) {

                a0 = *h; call inv; a2 = a0*a3;

                  a1 = *b-- - *b++; nop; nop;
                *a++ = a1 = a1 * a2;    

                a2 = *y - *y++; nop;
                a0 = *two;
                *c = a2 = a2 * a0;  /* a0 not updated yet */
                a1 = a3 * *h++;
                a2 = *b + a0 * *b++; nop; nop; nop;
                *c++ = a2 = *c - a2 * a1;

                }
        

        n = *input.n;  /* compute b and d */
        b = *input.b;
        y = *input.y;
        d = *input.d;

         half = HALF;

        i = n-2; do i+1

                { *b++ = a0 = *half * *b; *d++ = a0 = *y++; }



        /* Interpolate the spline. */

          x  = *input.x;
          y  = *input.y;
          h  = *input.h;
          a  = *input.a;
          b  = *input.b;
          c  = *input.c;
          d  = *input.d;
          xx = *input.xx;
          yy = *input.yy;
          nn = *input.nn;

        xdiv = input.xdiv; *xdiv = a0 = dsp( *xdiv);


        *yy++ = a1 = *y;
          *xx = a1 = *x++;

        for ( i = 1 ; i < nn ; i++ ) {

            *xx = a1 = a0 + *xx++; nop;


                a3 = a1 - *x; nop; nop; nop;

                    while ( age ) { /* inc to the next interval */

                        a += 4; b += 4; c += 4; d += 4; h += 4; x += 4;

                           a3 = a1 - *x; nop;
                        }


                /* Compute the cubic. */

                        a1 = *h + a3; nop; nop;
                        a2 = *b + a1 * *a; nop; nop;
                        a2 = *c + a2 * a1; nop; nop;
                *yy++ = a2 = *d + a2 * a1;

                }



        /* Convert the output back to ieee floats. */

        r1 = *input.n;  r2 = *input.h; call to_ieee;
        r1 = *input.n;  r2 = *input.a; call to_ieee;
        r1 = *input.n;  r2 = *input.b; call to_ieee;
        r1 = *input.n;  r2 = *input.c; call to_ieee;
        r1 = *input.n;  r2 = *input.d; call to_ieee;

        r1 = *input.nn; r2 = *input.xx; call to_ieee;
        r1 = *input.nn; r2 = *input.yy; call to_ieee;



        /* Restore the return address back into r14, and return. */

        sp -= 4;
        ra = *sp;

        return;

}


/* Wait state setting:
 *
 * Set the global wait state variable pcw_waits.
 *
 *
 */

constant

        A_0WS  = 0x0000,    /* 0 wait states for parition A */
        A_1WS  = 0x0002,    /* 1 wait states for parition A */
        A_2WS  = 0x0012,    /* 2 wait states for parition A */
        A_3WS  = 0x0022,    /* 3 wait states for parition A */

        B_0WS  = 0x0000,    /* 0 wait states for parition A */
        B_1WS  = 0x0001,    /* 1 wait states for parition A */
        B_2WS  = 0x0005,    /* 2 wait states for parition A */
        B_3WS  = 0x0009;    /* 3 wait states for parition A */

long
        pcw_waits = A_2WS|B_2WS,  /* default to 2 for A and B */

          pcwA[] = { A_0WS, A_1WS, A_2WS, A_3WS },
          pcwB[] = { B_0WS, B_1WS, B_2WS, B_3WS };


set_waits {

        struct input_array

                { long func, A, B; } input;

        register

                A       = r1,
                B       = r2,
                p       = r3,
                value   = r4;


        /* Set the wait states for memory partitions A and B. */

        A = *input.A; B = *input.B;

           A &= 0x3; A *= 4; A += pcwA; A = *A;
           B &= 0x3; B *= 4; B += pcwB; B = *B;
           
                p = pcw_waits; value = A|B; *p = value;

        return;

}



/* Subroutine call and return address stack:
 *
 * This stack is used by the functions above for saving and restoring the
 * the return address in r14, and is always pointed to by r13.
 *
 *
 */

long
        stack[100];




/* Marker indicating last memory location used for DSPMATH functions
 * in memory partition A.
 *
 */

constant

        end_of_functions = . ;  /* . = current assembler program location */






/* Data stack for piling up arrays in memory partition B.
 *
 * As noted in the comments at the top of this file, for older 640K boards
 * define DATA_STACK as 0x800000, or as . for newer boards.
 *
 *
 */

constant

         DATA_STACK     = . ;   /* . = end_of_functions for new boards  */


<DATA_STACK>

        long data_stack[];
