/* FILE: dspio.c                Copyright (c), Symmetric Research, 1999-2000
 *
 * This is the Linux version of the low level DSP32C support functions.  It
 * acts as a shell or wrapper around the Linux kernel mode device driver
 * dsp32c.sys that does all the actual work.
 * 
 * The function calls in this file agree with those in the Win9x, WinNT and 
 * DOS dspio.c.  Applications that work with those files are likely to work 
 * with this one with little modification.
 * 
 * However, in order for this Linux version to work, the underlying Linux
 * device driver dsp32ckd.o must be installed.  See the driver directory
 * for more details.
 * 
 *
 */


#include <stdio.h>
#include <errno.h>            /* errno global   */
#include <fcntl.h>            /* open  function */

#ifdef SROS_LINUX
#include <unistd.h>           /* close function */
#include <sys/ioctl.h>        /* ioctl function */
#endif




/* INCLUDE DSPIO.H:
 *
 * Note that for building the DLL library, the symbol _DLLEXPORT should be
 * defined on the compile command line.  See the makefile for details.
 *
 *
 *
 */

#include "dspio.h"
#include "dsp32ckd.h"   /* for device I/O control constants */







/* High and low word helper macros:
 *
 * Access the high and low 16 bit words of a 32 bit long or pointer.
 *
 *
 *
 */

#define HW( p)          ((unsigned short)(p>>16))
#define LW( p)          ((unsigned short)(p>>0))




/* Global variables
 *
 */

int    device_file       = 0;    /* Device driver file descriptor */
ULONG  last_driver_error = 0;




/************* START Linux specific functions **************/

/*  Deal with terminal attributes and keyboard entry
 *
 */

#ifdef SROS_LINUX

#include <unistd.h>          /* STDIN_FILENO define */
#include <termios.h>
#include <sys/stat.h>        /* fstat function */

int SetNonBlock( int value ) {

        int oldflags, result;

        oldflags = fcntl( STDIN_FILENO, F_GETFL, 0 );

        if (oldflags == -1)
                return( -1 );

        if (value != 0)
                oldflags |= O_NONBLOCK;
        else
                oldflags &= ~O_NONBLOCK;

       result = fcntl( STDIN_FILENO, F_SETFL, oldflags );
       return( result );
}

int  _commit( int fh ) {

        /* Pretend success */
        return( 0 );
}

int _filelength( int fh ) {

        struct stat buffer;

        if ( fstat( fh, &buffer ) == -1 ) {
                printf( "Could not get file stats\n" );
                return( -1 );
                }

        return( buffer.st_size );
}

#endif

/************* END Linux specific functions **************/





/* RELEASE REV INFO:
 *
 * Allow the user to determine the library rev they are working with.  See
 * dspio.h for more details.
 *
 */

VARTYPE( INT ) DSP32C_Rev = DSP32C_REV;

FUNCTYPE( INT ) DSP32C_GetRev( void ) {

        /* Return the library rev number. */

        return( DSP32C_Rev );
}






/* DSP32C DEVICE DRIVER ACCESS FUNCTIONS:
 *
 * These functions open and close the communication connection between
 * the DLL and the device driver.
 *
 * It is recommended that application programs begin by calling DSP32C_Open 
 * to initiate communication and end by calling DSP32C_Close to terminate
 * driver communication.
 *
 * However, to maintain compatibility with existing programs, DSP32C_Init
 * will automatically call DSP32C_Open if necessary.  Also, the functionality
 * of DSP32C_Close is automatically performed by the operating system when 
 * the application terminates.
 *
 */


FUNCTYPE( INT) DSP32C_Open( void ) {

        /* Only allow one open at a time */

        if (device_file != 0) {
                last_driver_error = -EBUSY;
                return( 1 );
                }

        
        /* Open device driver */

        device_file = open( DLL_DEVICE_NAME, O_RDWR );
        if (device_file < 0) {
                last_driver_error = errno;
                return( 0 );
                }

        last_driver_error = 0;


        /* Retrieve true values from device driver */

        DSP32C_GetBaseIo();


        return( 1 );
}

FUNCTYPE( INT) DSP32C_Close( void ) {

        /* Only close open drivers */

        if (device_file == 0) {
                last_driver_error = -EPERM;
                return( 1 );
                }

        /* Close driver */
        
        if ( close(device_file) < 0 ) {
                last_driver_error = errno;
                return( 0 );
                }


        /* Reset handle */
        
        device_file       = 0;
        last_driver_error = 0;

        return( 1 );
}


/* DEVICE DRIVER HELPER FUNCTION:
 *
 * This function calls ioctl to communicate information between
 * the DLL functions and the kernel mode device driver.
 *
 * After the device driver has been opened with the open function,
 * it is ready to provide its services.  Common services such as reading
 * and writing data are requested with the read and write
 * functions.  More specialized services are requested using the
 * ioctl function.
 *
 * The second argument to the ioctl function is an IO control
 * code that indicates which service is being requested.  The available
 * control codes are given in dsp32cbkd.h.
 *
 * The ioctl function allows small amounts of data to be 
 * exchanged between the caller and the device driver.  Argument three
 * is a pointer to the data buffer passed to the device driver.  The
 * size of the data must match the size defined for that IO control code
 * in dsp32ckd.h.
 *
 */

FUNCTYPE( INT) DevIoReadWrite( ULONG io_ctl_code, void *pdata ) {

        int  code;


        /* Device driver must be opened first */
        
        if (device_file == 0) {
                last_driver_error = -EPERM;
                return( 0 );
                }


        
        
        /* Request specialized services from device driver. */

        code = ioctl( device_file, io_ctl_code, (ULONG) pdata );

        
        /* Failure output */

        if (code != 0) {
                last_driver_error = errno;
                return( 0 );
                }
        
        last_driver_error = 0;

        return( 1 );
}












/* PC IO SPACE BASE ADDRESS:
 *
 *
 *
 *
 *
 *
 *
 */

#define DEFAULT_BASEIO          0x280

VARTYPE( INT) DSP32C_BaseIo  = 0x0000;


FUNCTYPE( INT) DSP32C_GetBaseIo( void) {

        PASSEDPARMS passed;


        /* Pass info to kernel mode driver if it is open */
        
        if (device_file != 0) {

                DevIoReadWrite( IOCTL_DSP32C_DSPGETBASE, &passed );

                DSP32C_BaseIo  = passed.baseio;
                DSP32C_BaseSel = passed.basesel;
                }


        return( DSP32C_BaseIo);

}
FUNCTYPE( VOID) DSP32C_SetBaseIo( unsigned NewBaseIo) {

        PASSEDPARMS passed;

        DSP32C_BaseIo = NewBaseIo;

        
        /* Pass info to kernel mode driver if it is open */
        
        if (device_file != 0) {

                passed.baseio  = DSP32C_BaseIo;
                passed.basesel = DSP32C_BaseSel;
        
                DevIoReadWrite( IOCTL_DSP32C_DSPSETBASE, &passed );
                }
}




/* DSPMUL CPU SELECTION:    (*** ONLY FOR DSPMUL MULTIPLE CPU BOARDS ***)
 *
 * For DSPMUL boards, which DSP32C cpu is currently connected to the ISA
 * bus is controlled by an additional on board PC IO space register.
 *
 * This byte wide register should be poked with the value of the desired
 * cpu.  DSP32C_SelectCpu MUST be called BEFORE DSP32C_Init to specify
 * which cpu will be initialized.  After initializing each cpu, it can
 * be called as needed to select a particular cpu for subsequent
 * operations.
 *
 *
 */

#define DEFAULT_BASESEL         0x2A0

VARTYPE( INT) DSP32C_BaseSel = 0x0000;


FUNCTYPE( INT) DSP32C_GetBaseSel( void) {

        PASSEDPARMS passed;


        /* Pass info to kernel mode driver if it is open */
        
        if (device_file != 0) {

                DevIoReadWrite( IOCTL_DSP32C_DSPGETBASE, &passed );

                DSP32C_BaseIo  = passed.baseio;
                DSP32C_BaseSel = passed.basesel;
                }


        return( DSP32C_BaseSel);

}
FUNCTYPE( VOID) DSP32C_SetBaseSel( unsigned NewBaseSel) {

        PASSEDPARMS passed;

        DSP32C_BaseSel = NewBaseSel;

        
        /* Pass info to kernel mode driver if it is open. */
        
        if (device_file != 0) {

                passed.baseio  = DSP32C_BaseIo;
                passed.basesel = DSP32C_BaseSel;
        
                DevIoReadWrite( IOCTL_DSP32C_DSPSETBASE, &passed );
                }
}



FUNCTYPE( VOID) DSP32C_SelectCpu( unsigned n) {

        /* Ensure that kernel mode device driver is open. */

        if (device_file != 0 && DSP32C_Open() != 0) {
                last_driver_error = -EPERM;
                return;
                }


        /* Set the DSPMUL cpu selection register. */

        DevIoReadWrite( IOCTL_DSP32C_DSPSELOUT, &n );

}




/* DSP32C PIO REGISTER OFFSETS AND READ/WRITE FUNCTIONS:
 *
 *  NOTE: These ** MUST ** match the defines used in dsp32ckd.c.
 *
 *
 *
 *
 */

#define PCRl            (0x07*2)    /* DSP32 mode PCR low              */
#define PCRh            (0x0A*2)    /* DSP32C 8 bit mode PCR high      */

#define PAR             (0x00*2)    /* DSP32C 16 bit mode registers    */
#define PDR             (0x02*2)
#define EMR             (0x04*2)
#define ESR             (0x06*2)
#define PCR             (0x07*2)
#define PIR             (0x08*2)
#define PARE            (0x0B*2)
#define PDR2            (0x0C*2)







static int ReadPio( int reg_offset) {

        PASSEDPARMS passed;
        int         value;


        /* Read a PIO register. */

        passed.reg_offset = reg_offset;
        passed.value      = 0;
        
        DevIoReadWrite( IOCTL_DSP32C_DSPIN, &passed );

        value = (int)passed.value;

        return( value);
}

static void WritePio( int reg_offset, int value) {

        PASSEDPARMS passed;

        /* Write a PIO register. */

        passed.reg_offset = reg_offset;
        passed.value      = (short)value;
        
        DevIoReadWrite( IOCTL_DSP32C_DSPOUT, &passed );

}















/* PROCESSOR INITIALIZATION:
 *
 * Note that DSP32C_BaseIo must be set before calling these functions!
 *
 *
 * In the past, setting DSP32C_BaseIo and DSP32C_BaseSel with DOS
 * environment space variables has been allowed.  This is supported here
 * with the C runtime getenv function.  THIS IS THE ONLY PLACE IN
 * DSPIO.C THAT THE C RUNTIME LIBRARY IS CALLED.
 *
 * If you wish to avoid the C runtime, compile without the calls to
 * getenv, and set up DSP32C_BaseIo and DSP32C_BaseSel on your own.
 *
 *
 * Actually, all this is meaningless with the current NT device driver
 * since the values of DSP32C_BaseIo and DSP32C_BaseSel are controlled
 * ONLY by the values specified at the time the driver is installed.
 * The DSP32C_GetBaseIo and DSP32C_GetBaseSel functions will return
 * this value.  You can also see them by looking at the DSP32C load
 * message in the System EventLog.
 *
 */


FUNCTYPE( INT) DSP32C_Init( void) {

        /* Halt the DSP32C and configure the PIO port.
         *
         * The sequence of step are:
         *
         *   1.  set REGMAP for DSP32C PAB mode
         *   2.  convert to a 16 bit interface by setting PIO16
         *   3.  enable DMA and AUTO PIO data transfer modes
         *
         *
         */

        /* Ensure that kernel mode device driver is open. */

        if (device_file == 0) {

                if ( !DSP32C_Open() ) {
                        last_driver_error = -EPERM;
                        return(0);
                        }
                }


        /* Retrieve actual BaseIo and BaseSel values from device driver. */

        DSP32C_GetBaseIo();
        

        WritePio( PCRl, 0x02);         /* set REGMAP -> DSP32C PAB map */
        WritePio( PCRh, 0x02);         /* set PIO16 -> 16 bit interface */

        WritePio( PCR, 0x21A);         /* set DMA and AUTO */




        /* Do a presence test, checking the PCR register value. */

        if ( (ReadPio( PCR) & 0x71FL) != 0x21A )
                return( 0);



        /* Return success. */

        return( 1);


}










/* EXECUTION CONTROL:
 *
 *
 *
 *
 *
 *
 *
 */

FUNCTYPE( VOID) DSP32C_Go( void) {


        /* Reset the DSP32C and start execution from location 0.
         *
         * NOTE:
         *
         * As with DSP32C_Init, the PIO port must be reprogrammed for the 16
         * bit DMA mode after reset.
         *
         *
         */


        /* Toggle PCR[0], initiating reset and execution. */

        WritePio( PCRl, 0x00);
        WritePio( PCRl, 0x01);



        /* Reprogram the PIO port WHILE executing. */

        WritePio( PCRl, 0x03);         /* set REGMAP for DSP32C PAB map */
        WritePio( PCRh, 0x02);         /* set PIO16 for 16 bit interface */

        WritePio( PCR, 0x21B);         /* set DMA and AUTO */


}
FUNCTYPE( VOID) DSP32C_Halt( void) {

        int value;


        /* Halt the DSP32C by clearing PCR[0]. */

        value = ReadPio( PCR);

            value &= (~0x1);


        WritePio( PCR, value);

}










/* PIO REGISTER FUNCTIONS:
 *
 * These functions read and write the DSP32C PIO registers.
 *
 *
 *
 */

FUNCTYPE( INT) DSP32C_Pcr( void) {

        return( ReadPio( PCR));

}
FUNCTYPE( INT) DSP32C_Pir( void) {

        return( ReadPio( PIR));

}
FUNCTYPE( INT) DSP32C_Esr( void) {

        return( ReadPio( ESR));

}
FUNCTYPE( INT) DSP32C_Emr( void) {

        return( ReadPio( EMR));

}
FUNCTYPE( VOID) DSP32C_SetPir( unsigned value) {


        /* Set the PIR register to the specified value from the PC. */

        WritePio( PIR, value);


}
FUNCTYPE( VOID) DSP32C_SetEni( unsigned value) {

#define ENI     0x0004

        unsigned pcr;



        /* Set the value of the PCR[2] ENI bit.
         *
         * This bit enables or disables the DSP32C PIF pin which is connected
         * to the PC interrupt header on the board.  Value = 1 enables the
         * pin, value = 0 disables it.
         *
         * Note that the eni bit is automatically cleared by a reset of the
         * DSP32C, and must be reenabled after DSP32C_Init or Go if interrupts
         * are being used.
         *
         */

        pcr = ReadPio( PCR);

            if ( value == 1 )

                    pcr |= ENI;

            else
                    pcr &= ~ENI;



        WritePio( PCR, pcr);


}















/* DATA TRANSFER AND PROGRAM DOWNLOADING:
 *
 *
 *
 *
 *
 */

FUNCTYPE( VOID) DSP32C_Get(unsigned long src,void *dest,unsigned long nbytes) {


        PASSEDBUFFPARMS parm;


        /* Set up parameters to be passed to driver. */

        parm.dspaddr    = src;
        parm.count      = nbytes;
        parm.buffer     = dest;
        


        /* Move data from the DSP32C to the PC, where:
         *
         *      src = DSP32C 24 bit source address
         *     dest = PC destination address
         *
         *
         */

        DevIoReadWrite( IOCTL_DSP32C_DSPINSW, &parm );

}


FUNCTYPE( VOID) DSP32C_Put(void *src,unsigned long dest,unsigned long nbytes) {


        PASSEDBUFFPARMS parm;


        /* Set up parameters to be passed to driver. */

        parm.dspaddr    = dest;
        parm.count      = nbytes;
        parm.buffer     = src;
        


        /* Move data from the PC to the DSP32C, where:
         *
         *      src = PC source address
         *     dest = DSP32C 24 bit destination address
         *
         *
         */

        DevIoReadWrite( IOCTL_DSP32C_DSPOUTSW, &parm );

}


FUNCTYPE( INT) DSP32C_SrHexLoad( unsigned long *input) {

        unsigned long dsp_start, nbytes;



        /* Download a DSPASM .dsh hex input array. */

        DSP32C_Halt();



        /* Check the leading array id. */

        if ( *input++ != 0x52535F5FL /* = "__SR */ )

                return( 0);



        /* Download records. */

        while ( *input != 0xFFFFFFFFL ) {

                dsp_start = *input++;
                   nbytes = *input++;

                   DSP32C_Put( (void *)input, dsp_start, nbytes);

                      ((char *)input) += nbytes;

                }



        /* Return success. */

        return( 1);

}







/* PUT LONG:
 *
 * This function writes a single 32 bit value to the DSP32C using the PIO
 * port 32 bit DMA mode.  The 16 bit PIO DMA autoinc mode generally used
 * by DSP32C_Get and Put is saved and restored on entry and exit.
 * 
 * This function is useful because it holds the PIO port off from
 * writing until it has a full 32 bits.  This allows the PC to poke
 * instructions and data onto the board while competing with DSP32C
 * execution, and not risk fetching problems as in the case of two 16
 * bit sequential writes.
 * 
 * The disadvantage of using this function is that it does not use the
 * outsw instruction like DSP32_Put.  In general, this function should
 * only be used for specialized applications, like breakpoint execution
 * in the DSPMON monitor.
 * 
 * As with the (PAR,PARE) registers, the order in which (PDR2,PDR) is
 * written is VERY important.  See the AT&T Information manual for
 * details.
 *
 *
 */

FUNCTYPE( VOID) DSP32C_PutLong( unsigned long value, unsigned long addr) {

        unsigned  save, temp;


        PASSEDBUFFPARMS parm;


        /* Set up parameters to be passed to driver.  Ignore the
         * "count" name.  It is really being used to send the value
         * in this case.
         */


        parm.dspaddr = addr;
        parm.count   = value;
        


        /* Save the current DSP32C PCR value. */

        save = ReadPio( PCR);



        /* Program the PCR for a 32 bit DMA transfer. */

        temp = save;

            temp &= 0x0465;             /* mask off bits not to change */
            temp |= 0x031A;             /* set the 32 bit DMA mode */

            WritePio( PCR, temp);



        /* Transfer the long value. */

        DevIoReadWrite( IOCTL_DSP32C_DSPOUTLONG, &parm );



        /* Restore the PCR. */

        WritePio( PCR, save);

}

